"""Module containing classes and functions used for styling plotly figures.

Contains:
- LineStyle: Dataclass containing information on how to style a line on a line
    chart.
"""
from .line_style import LineStyle



"""enable_basic_auth"""

import os
import dash
from flask_basicauth import BasicAuth


def enable_basic_auth(dash_app: dash.Dash):
    """Turn on basic auth for the dash app if required"""
    if os.environ.get("STAGE") == "production":
        if os.environ.get("APP_USERNAME") and os.environ.get("APP_PASSWORD"):
            server = dash_app.server

            # Forcing the auth to prompt on any page if it hasn't already been done.
            server.config["BASIC_AUTH_FORCE"] = True

            server.config["BASIC_AUTH_USERNAME"] = os.environ.get("APP_USERNAME")
            server.config["BASIC_AUTH_PASSWORD"] = os.environ.get("APP_PASSWORD")
            BasicAuth(server)
        else:
            print("failed to set up basic_auth")



"""download_button"""
import warnings
from dash import html

from gov_uk_dashboards.constants import DOWNLOAD_BUTTON_CLASSES


def download_button(button_text: str, button_id: str = "download-button"):
    """
    Return a download button which is aligned to the right

    Args:
    button_text (str): The text to display on the button.
    button_id: (str, Optional) = id for dropdown, default to "download-button"
    """
    warnings.warn(
        "Note there is an alternative function to download_button() called "
        "create_download_button_with_icon() which includes a download icon and improved styling.",
        Warning,
        stacklevel=2,
    )

    return html.Div(
        [
            html.Button(
                button_text,
                id=button_id,
                n_clicks=0,
                className="govuk-button govuk-button--secondary",
            ),
        ],
        className="govuk-button-group",
        style={"float": "right"},
    )


def create_download_button_with_icon(
    button_text: str, button_id_name: str
) -> html.Button:
    """Create a download button with icon, aligned to the left.

    Parameters:
    - button_id_name (str): A unique identifier for the button.

    Returns:
    - html.Button: Download button.
    """
    download_type = button_text.lower().replace(" ", "-")
    return html.Button(
        [
            html.Div("", className="download-icon"),
            button_text,
        ],
        id={
            "download-type": download_type,
            "name": button_id_name,
        },
        n_clicks=0,
        className=DOWNLOAD_BUTTON_CLASSES,
        type="submit",
        style={
            "display": "flex",
            "alignItems": "center",
            "gap": "8px",
        },
    )



"""phase_banner"""
from dash import html


def phase_banner_with_feedback(
    phase: str,
    feedback_link: str,
    link_id: str = "feedback-link",
    link_target: str = "_self",
):
    """Creates a phase banner with a feedback link, which can be specified.
    The id of the 'feedback' link can be set to allow for targeting with callbacks.
    By default the link will open on the same window, to open in a new window
    set link_target to "_blank"
    The phase banner is set out here as part of the Gov.UK design system:
    https://design-system.service.gov.uk/components/phase-banner/"""
    return html.Div(
        [
            html.P(
                [
                    html.Strong(
                        [phase],
                        className="govuk-tag govuk-phase-banner__content__tag",
                    ),
                    html.Span(
                        [
                            "This is a new service - your ",
                            html.A(
                                ["feedback"],
                                href=feedback_link,
                                className="govuk-link",
                                id=link_id,
                                target=link_target,
                                rel="noopener noreferrer"
                                if link_target == "_blank"
                                else "",  # conditional attribute for security
                            ),
                            " will help us to improve it.",
                        ],
                        className="govuk-phase-banner__text",
                    ),
                ],
                className="govuk-phase-banner__content",
            )
        ],
        className="govuk-phase-banner",
    )



"""footer"""
from typing import Optional
from dash import html


def footer(footer_links: Optional[list[any]]):
    """
    HTML component for a Gov.UK standard footer.

    The footer provides copyright, licensing and other information about your
    service and department.

    Part of the Gov.UK Design System:
    https://design-system.service.gov.uk/components/footer/

    Returns:
        html.Footer: A footer element with copyright information.
    """
    return html.Footer(
        html.Div(
            html.Div(
                [
                    html.Div(
                        [
                            html.H2("Support Links", className="govuk-visually-hidden")
                            if footer_links
                            else None,
                            html.Ul(
                                children=[
                                    html.Li(
                                        item, className="govuk-footer__inline-list-item"
                                    )
                                    for item in footer_links
                                ],
                                className="govuk-footer__inline-list govuk-!-display-none-print",
                            )
                            if footer_links
                            else None,
                            html.Span(
                                [
                                    "All content is available under the ",
                                    html.A(
                                        "Open Government Licence v3.0",
                                        rel="license",
                                        href="https://www.nationalarchives.gov.uk/doc/"
                                        "open-government-licence/version/3/",
                                        className="govuk-footer__link",
                                        target="_blank",
                                    ),
                                    ", except where otherwise stated",
                                ],
                                className="govuk-footer__licence-description",
                            ),
                        ],
                        className="govuk-footer__meta-item govuk-footer__meta-item--grow",
                    ),
                    html.Div(
                        html.A(
                            "© Crown copyright",
                            className="govuk-footer__link pyvis-govuk-footer__copyright-logo",
                            href="https://www.nationalarchives.gov.uk/information-management/"
                            "re-using-public-sector-information/uk-government-licensing-framework/"
                            "crown-copyright/",
                            target="_blank",
                        ),
                        className="govuk-footer__meta-item",
                    ),
                ],
                className="govuk-footer__meta",
            ),
            className="govuk-width-container",
        ),
        className="govuk-footer",
        role="contentinfo",
    )



"""Module containing Dash components for building GOV.UK Design System dashboards.

See: https://design-system.service.gov.uk/components/

Contains:
- message_banner: Return a changelog banner to be used to communicate to the
    user when the dashboard was last updated.
- captioned_figure: Return figure with attached captions that can be read
    by a screen reader.
- card: Return a rectangle with a grey background. Mostly used to wrap
    individual visualisations.
- card_full_width: Return a card with a grey background that fits the full
    width of its parent container using CSS flexbox.
- empty_card: Return an empty card that is hidden to help keep alignment with
    rows and columns.
- collapsible_panel: Return a component that allows the user to open and close
    a collapsible panel containing child components.
- dashboard_container: Return a HTML wrapper for a whole dashboard.
- details: Return HTML component for showing a summary expandable with more
    information beneath.
- filter_panel: Return a card with a title that allows the user to select and
    filter metrics on the dashboard.
- hidden_filter: Return an empty, invisible HTML element that stands in for a
    filter component.
- footer: Return a component for a Gov.UK standard footer.
- graph: Takes a Plotly Figure and returns a responsive Dash graph with useful
    defaults.
- heading1: Return a H1 Dash component with gov.uk styling
- heading2: Return a H1 Dash component with gov.uk styling
- heading3: Return a H1 Dash component with gov.uk styling
    - HeadingSizes: Enum of sizes that can be used to manually adjust
        text size for heading function returns.
- html_list: Return either a <ul> or <ol> component, with the children set as
    a list of <li> components matching the list_items provided.
- key_value_pair: Return an element that displays a value (such as a metric)
    labelled by a key (such as the name of the metric).
- main_content: Return wrapper for the main content of the dashboard,
    containing visualisations.
- navbar: Return a navigation bar for switching between dashboards.
    - navbar_link: Return a link for use with the navbar.
    - navbar_link_active: Return a link that appears highlighted, suggesting to
        the user that they are already viewing the linked dashboard.
- side_navbar: Return a navigation bar for switching between dashboards.
    - side_navbar_link: Return a link for use with the side_navbar.
    - side_navbar_link_active: Return a link that appears highlighted,
        suggesting to the user that they are already viewing the linked
        dashboard.
- no_data_message: Return a list of strings with a message for when
    selection does not provide data.
- paragraph: Return a formatted <p> html component with the children provided.
    - ParagraphSizes: Enum for use with paragraph to specify the text size.
- phase_banner_with_feedback: Return a phase banner with a feedback link,
    which can be specified.
- row_component: Returns a horizontal row used to contain cards.
- table_from_dataframe: Return a html table created from a pandas DataFrame.
- tooltip_title: Return a tooltip component for explaining details.
- format_visualisation_commentary: Return paragraph styling commentary.
- format_visualisation_title: Return a default formatted title for a
    visualisation.
- apply_and_reset_filters_buttons: Add apply filters and reset filters buttons, whcih are aligned to
the right
- add_filter_button: Return a 'Compare to additional authority' button which is aligned to the right
"""

from .banners import message_banner
from ..plotly.captioned_figure import captioned_figure
from .card import card, empty_card
from .card_full_width import card_full_width
from .apply_and_reset_filters_buttons import apply_and_reset_filters_buttons
from .download_button import download_button
from .collapsible_panel import collapsible_panel
from .dashboard_container import dashboard_container
from .details import details
from .filter_panel import filter_panel, hidden_filter
from .footer import footer
from .graph import graph
from .heading import heading1, heading2, heading3, HeadingSizes
from .html_list import html_list
from .key_value_pair import key_value_pair
from .main_content import main_content
from .navbar import navbar, navbar_link, navbar_link_active
from .no_data_message import no_data_message
from .paragraph import paragraph, ParagraphSizes
from .phase_banner import phase_banner_with_feedback
from .row_component import row_component
from .side_navbar import side_navbar
from .table import table_from_dataframe
from .tooltip_title import tooltip_title
from .visualisation_commentary import format_visualisation_commentary
from .visualisation_title import format_visualisation_title
from .comparison_la_filter_button import add_filter_button



"""get_time_series_chart function"""

from datetime import datetime, date
import json
from typing import Optional
from dateutil.relativedelta import relativedelta
from dash import html
import polars as pl

import plotly.graph_objects as go

from gov_uk_dashboards.constants import (
    CHART_LABEL_FONT_SIZE,
    CUSTOM_DATA,
    DATE_VALID,
    HOVER_TEXT_HEADERS,
    LEGEND_SPACING,
    MAIN_TITLE,
    REMOVE_INITIAL_MARKER,
    SUBTITLE,
    YEAR,
)
from gov_uk_dashboards.colours import AFAccessibleColours
from gov_uk_dashboards.components.helpers.display_chart_or_table_with_header import (
    display_chart_or_table_with_header,
)
from gov_uk_dashboards.components.helpers.generate_dash_graph_from_figure import (
    generate_dash_graph_from_figure,
)
from gov_uk_dashboards.components.helpers.plotting_helper_functions import (
    get_legend_configuration,
    get_rgba_from_hex_colour_and_alpha,
)
from gov_uk_dashboards.components.helpers.get_chart_for_download import (
    get_chart_for_download,
)
from gov_uk_dashboards.components.helpers.update_layout_bgcolor_margin import (
    update_layout_bgcolor_margin,
)
from gov_uk_dashboards.components.plotly.enums import (
    HoverDataByTrace,
    TitleDataStructure,
    XAxisFormat,
)
from gov_uk_dashboards.lib.datetime_functions.datetime_functions import (
    convert_financial_quarter_to_financial_quarter_text,
    replace_jun_jul_month_abbreviations,
)


class TimeSeriesChart:
    """Class for use in generating time series charts."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-positional-arguments

    def __init__(
        self,
        title_data: TitleDataStructure,
        y_axis_column: str,
        hover_data: HoverDataByTrace,
        filtered_df: pl.DataFrame,
        trace_name_list: list[str],
        hover_data_for_traces_with_different_hover_for_last_point: Optional[
            HoverDataByTrace
        ] = None,
        legend_dict: dict[str, str] = None,
        trace_name_column: Optional[str] = None,
        xaxis_tick_text_format: XAxisFormat = XAxisFormat.YEAR.value,
        verticle_line_x_value_and_name: tuple = None,
        filled_traces_dict: dict[str] = None,
        trace_names_to_prevent_hover_of_first_point_list=None,
        x_axis_column=DATE_VALID,
        x_unified_hovermode: Optional[bool] = False,
        x_hoverformat: Optional[str] = None,
        x_unified_hovertemplate: Optional[str] = None,
        x_axis_title: Optional[str] = None,
        download_chart_button_id: Optional[str] = None,
        download_data_button_id: Optional[str] = None,
        number_of_traces_colour_shift_dict: Optional[dict] = None,
        additional_line: Optional[dict] = None,
        hover_distance: Optional[int] = 1,
        footnote: Optional[str] = None,
    ):
        self.title_data = title_data
        self.y_axis_column = y_axis_column
        self.hover_data = hover_data
        self.hover_data_for_traces_with_different_hover_for_last_point = (
            hover_data_for_traces_with_different_hover_for_last_point
        )
        self.filtered_df = filtered_df
        self.trace_name_list = trace_name_list
        self.legend_dict = legend_dict
        self.trace_name_column = trace_name_column
        self.xaxis_tick_text_format = xaxis_tick_text_format
        self.verticle_line_x_value_and_name = verticle_line_x_value_and_name
        self.filled_traces_dict = filled_traces_dict
        self.trace_names_to_prevent_hover_of_first_point_list = (
            trace_names_to_prevent_hover_of_first_point_list
        )
        self.x_axis_column = x_axis_column
        self.x_unified_hovermode = x_unified_hovermode
        self.x_hoverformat = x_hoverformat
        self.x_unified_hovertemplate = x_unified_hovertemplate
        self.x_axis_title = x_axis_title
        self.download_chart_button_id = download_chart_button_id
        self.download_data_button_id = download_data_button_id
        self.markers = ["square", "diamond", "circle", "triangle-up"]
        self.number_of_traces_colour_shift_dict = number_of_traces_colour_shift_dict
        self.additional_line = additional_line
        self.hover_distance = hover_distance
        self.colour_list = self._get_colour_list()
        self.fig = self.create_time_series_chart()
        self.footnote = footnote

    def get_time_series_chart(self) -> html.Div:
        """Creates and returns time series chart for display on application.
        Returns:
            html.Div: Styled div containing title, subtile and chart.
        """
        # pylint: disable=duplicate-code
        graph_name = self.title_data[MAIN_TITLE].replace(" ", "-")
        return display_chart_or_table_with_header(
            generate_dash_graph_from_figure(
                self.fig,
                graph_name,
                class_name="default-cursor-graph non-interactive-legend-cursor",
            ),
            self.title_data[MAIN_TITLE],
            self.title_data[SUBTITLE],
            self.download_chart_button_id,
            self.download_data_button_id,
            self.footnote,
        )

    def is_json_serializable(self, value):
        "Determines whether value can be converted to json format."
        # pylint: disable=duplicate-code
        try:
            json.dumps(value)
            return True
        except (TypeError, OverflowError):
            return False

    def to_dict(self):
        "Converts class attributes to json format."
        # pylint: disable=duplicate-code
        result = {}
        for k, v in self.__dict__.items():
            if self.is_json_serializable(v):
                result[k] = v
            elif isinstance(v, pl.DataFrame):
                result[k] = {"_type": "polars_df", "data": v.to_dicts()}
            elif isinstance(v, pl.Series):
                result[k] = {"_type": "series", "data": v.to_list()}
            else:
                result[k] = f"<<non-serializable: {type(v).__name__}>>"
        return result

    @classmethod
    def from_dict(cls, data):
        "Creates a class instance from dict of attributes."
        restored = {}
        for k, v in data.items():
            if k in ["markers", "colour_list", "fig"]:
                continue
            if isinstance(v, dict) and "_type" in v:
                if v["_type"] == "polars_df":
                    restored[k] = pl.DataFrame(v["data"])
                elif v["_type"] == "polars_series":
                    restored[k] = pl.Series(v["data"])
                else:
                    # Fallback for unknown _type
                    restored[k] = v.get("data", None)
            else:
                restored[k] = v
        return cls(**restored)

    def get_time_series_chart_for_download(self):
        """Return fig with title and subtitle for download as png"""
        return get_chart_for_download(self, self.create_time_series_chart())

    def create_time_series_chart(
        self,
    ):
        """generates a time series chart"""
        # pylint: disable=too-many-locals
        # pylint: disable=duplicate-code

        if not self.x_unified_hovermode and self.x_hoverformat is not None:
            raise ValueError(
                "x_hoverformat can only be specified if x_unified_hovermode is True"
            )

        fig = go.Figure()
        if self.additional_line:
            x_0 = self.additional_line["x0"]
            y_0 = self.additional_line["y0"]
            x_1 = self.additional_line["x1"]
            y_1 = self.additional_line["y1"]
            line_color = self.additional_line["color"]
            trace_connector = go.Scatter(
                x=[x_0, x_1],
                y=[y_0, y_1],
                mode="lines",
                name="Transition",
                line={"color": line_color},
                hoverinfo="skip",  # 👈 This line disables hover for this trace
                showlegend=False,  # Optional: hide it from legend too
                legendgroup=self.additional_line["legend_group"],
            )
            fig.add_trace(trace_connector)
        # pylint: disable=unused-variable
        for i, (df, trace_name, colour, marker,) in enumerate(
            zip(
                self._get_df_list_for_time_series(),
                self.trace_name_list,
                self.colour_list,
                self.markers,
            )
        ):
            if REMOVE_INITIAL_MARKER in df.columns and True in df.get_column(
                REMOVE_INITIAL_MARKER
            ):
                marker_sizes = [0] + [12] * (len(df.with_row_count()) - 1)
            else:
                marker_sizes = [12] * (len(df.with_row_count()))
            legendgroup = self._get_legend_group(df)
            fig.add_trace(
                self.create_time_series_trace(
                    df.sort(self.x_axis_column),
                    trace_name,
                    line_style={"dash": "solid", "color": colour},
                    marker={"symbol": marker, "size": marker_sizes, "opacity": 1},
                    legendgroup=legendgroup,
                ),
            )

        if self.filled_traces_dict:
            fill_df = self.filtered_df.filter(
                pl.col(self.trace_name_column).is_in(
                    [
                        self.filled_traces_dict["upper"],
                        self.filled_traces_dict["lower"],
                    ]
                )
            ).sort(self.x_axis_column)
            x_series = fill_df[self.x_axis_column].unique().sort().to_list()
            upper_trace = self.filled_traces_dict["upper"]
            lower_trace = self.filled_traces_dict["lower"]
            y_upper = fill_df.filter(pl.col(self.trace_name_column) == upper_trace)[
                self.y_axis_column
            ].to_list()
            y_lower = fill_df.filter(pl.col(self.trace_name_column) == lower_trace)[
                self.y_axis_column
            ].to_list()
            y_upper_formatted_value = fill_df.filter(
                pl.col(self.trace_name_column) == upper_trace
            )["FORMATTED_VALUE"].to_list()
            y_lower_formatted_value = fill_df.filter(
                pl.col(self.trace_name_column) == lower_trace
            )["FORMATTED_VALUE"].to_list()
            formatted_dates = fill_df.filter(
                pl.col(self.trace_name_column) == upper_trace
            )["FORMATTED_DATE"].to_list()
            hover_text = [
                f"{self.filled_traces_dict['name']} - {date}: "
                f"{low_value} - {high_value}<extra></extra>"
                for low_value, high_value, date in zip(
                    y_lower_formatted_value,
                    y_upper_formatted_value,
                    formatted_dates,
                )
            ]

            hover_text_full = hover_text + hover_text[::-1]
            legendgroup = self._get_legend_group(fill_df)
            fig.add_trace(
                go.Scatter(
                    x=x_series + x_series[::-1],
                    y=y_upper + y_lower[::-1],
                    fill="toself",
                    fillcolor=get_rgba_from_hex_colour_and_alpha(
                        AFAccessibleColours.TURQUOISE.value, alpha=0.2
                    ),
                    line={"color": "rgba(255,255,255,0)"},
                    name=self.filled_traces_dict["name"] + LEGEND_SPACING,
                    hovertemplate=hover_text_full,
                    hoveron="points",
                    legendgroup=legendgroup,
                )
            )
        self._format_x_axis(fig)

        # if self.average_increment_for_average_trace is not None:
        #     trace_name = LINEAR_TRAJECTORY
        #     dates = [
        #         datetime(2024, 7, 9) + relativedelta(months=1 * i)
        #         for i in range(len(tick_values))
        #     ]
        #     values = [
        #         self.average_increment_for_average_trace * i for i in range(len(dates))
        #     ]
        #     fig.add_trace(
        #         go.Scatter(
        #             x=dates,
        #             y=values,
        #             mode="lines",
        #             name=trace_name,
        #             line={"dash": "dash", "color": "darkgray", "width": 2},
        #             hoverinfo="skip",
        #         )
        #     )
        if self.verticle_line_x_value_and_name is not None:

            fig.add_vline(
                x=self.verticle_line_x_value_and_name[0],
                line_width=2,
                line_dash="dash",
                line_color="#b3b3b3",
            )
            fig.add_annotation(
                x=self.verticle_line_x_value_and_name[0],
                yref="paper",
                y=0.9,
                text=self.verticle_line_x_value_and_name[1],
                showarrow=False,
                font={"color": "#414042", "size": 16},
                xanchor="left",
                yanchor="bottom",
            )

        y_range = [0, self._get_y_axis_range_max()]

        fig.update_yaxes(rangemode="tozero", showgrid=True, range=y_range)
        update_layout_bgcolor_margin(fig, "#FFFFFF")

        if self.x_axis_title is not None:
            fig.add_annotation(
                xref="x domain",
                yref="y domain",
                x=1,
                y=-0.2,
                text=self.x_axis_title,
                showarrow=False,
                font={"size": 16},
            )

        fig.update_layout(
            legend=get_legend_configuration(),
            font={"size": CHART_LABEL_FONT_SIZE},
            yaxis_tickformat=",",
            hovermode="x unified" if self.x_unified_hovermode is True else "closest",
            hoverdistance=self.hover_distance,  # Increase distance to simulate hover 'always on'
        )
        return fig

    def _get_legend_group(self, df):
        if "legend_group" in df.columns and len(df) > 0:
            value = df["legend_group"][0]
            legendgroup = value if value is not None else None
        else:
            legendgroup = None
        return legendgroup

    # pylint: disable=duplicate-code
    def _format_x_axis(self, fig):
        tick_text, tick_values, range_x = self._get_x_axis_content()

        fig.update_xaxes(
            tickvals=tick_values,
            ticktext=tick_text,
            tickmode="array",
            range=range_x,
            hoverformat=self.x_hoverformat,
        )

    def create_time_series_trace(
        self,
        df: pl.DataFrame,
        trace_name: str,
        line_style: dict[str, str],
        marker: dict[str, str],
        legendgroup: str,
    ):
        """Creates a trace for the plot.
        Args:
            df (pl.DataFrame): Dataframe to use to create trace. Must contain "Date valid" column,
            y_value column and columns defined in self.hover_data[CUSTOM_DATA].
            trace_name (str): Name of trace.
            line_style (dict[str, str]): Properties for line_style parameter.
            marker (dict[str,str]): Properties for marker parameter.
            legendgroup (str): Name to group by in legend,
        """
        return go.Scatter(
            x=df[self.x_axis_column],
            y=df[self.y_axis_column],
            line=line_style,
            name=self._get_trace_name(trace_name) + LEGEND_SPACING,
            hovertemplate=self._get_hover_template(df, trace_name),
            customdata=self._get_custom_data(df, trace_name),
            marker=marker,
            hoverlabel=None,
            showlegend=(
                trace_name in self.legend_dict if self.legend_dict is not None else True
            ),
            legendgroup=legendgroup,
        )

    def _get_hover_template(self, df, trace_name):
        return [
            (
                ""
                if i == 0
                and self.trace_names_to_prevent_hover_of_first_point_list is not None
                and trace_name in self.trace_names_to_prevent_hover_of_first_point_list
                else self._get_custom_hover_template(i, df, trace_name)
            )
            for i in range(df.shape[0])  # the number of rows in df
        ]

    def _get_custom_hover_template(self, i, df, trace_name):
        if self.x_unified_hovermode is True:
            if self.x_unified_hovertemplate is not None:
                return self.x_unified_hovertemplate.format(trace_name=trace_name)
            return f"{trace_name}: " + "%{customdata[0]}<extra></extra>"

        hover_text_headers = self.hover_data[trace_name][HOVER_TEXT_HEADERS]
        if (
            self.hover_data_for_traces_with_different_hover_for_last_point is not None
            and trace_name
            in self.hover_data_for_traces_with_different_hover_for_last_point
            and i == df.shape[0] - 1
        ):
            hover_text_headers = (
                self.hover_data_for_traces_with_different_hover_for_last_point[
                    trace_name
                ][HOVER_TEXT_HEADERS]
            )
        # pylint: disable=duplicate-code

        return (
            f"{trace_name}<br>"
            f"{hover_text_headers[0]}"
            ": %{customdata[0]}<br>"
            f"{hover_text_headers[1]}"
            ": %{customdata[1]}<extra></extra>"
        )

    def _get_custom_data(self, df, trace_name):
        # For last points of trace_name in [], we want different custom data.
        customdata = df[self.hover_data[trace_name][CUSTOM_DATA]]
        if (
            self.hover_data_for_traces_with_different_hover_for_last_point is not None
            and trace_name
            in self.hover_data_for_traces_with_different_hover_for_last_point
        ):
            customdata = [
                (
                    [df[col][i] for col in self.hover_data[trace_name][CUSTOM_DATA]]
                    if i < df.shape[0] - 1
                    else [
                        df[col][i]
                        for col in self.hover_data_for_traces_with_different_hover_for_last_point[
                            trace_name
                        ][
                            CUSTOM_DATA
                        ]
                    ]
                )  # Use different columns for the last point
                for i in range(df.shape[0])
            ]
        return customdata

    def _get_trace_name(self, trace_name):
        if self.legend_dict is not None and trace_name in self.legend_dict:
            return self.legend_dict[trace_name]
        return trace_name

    def _get_x_axis_content(self):
        """Generates tick text and values for the x-axis based on the unique years calculated from
        the DATE_VALID column in the dataframe.
        Returns:
            tuple: A tuple containing tick_text, tick_values and range_x.
        """
        if self.xaxis_tick_text_format == XAxisFormat.YEAR.value:
            df_with_year_column = self.filtered_df.with_columns(
                pl.col(self.x_axis_column)
                .str.strptime(pl.Date, "%Y-%m-%d")
                .dt.year()
                .alias(YEAR)
            )
            date_list = df_with_year_column[self.x_axis_column].unique().to_list()
            min_date_string = min(date_list)
            max_date_string = max(date_list)
            min_datetime = datetime.strptime(min_date_string, "%Y-%m-%d")
            max_datetime = datetime.strptime(max_date_string, "%Y-%m-%d")
            year_list = df_with_year_column[YEAR].unique().to_list()
            tick_text = list(range(min(year_list) - 1, max(year_list) + 2))
            tick_values = [date(year, 1, 1) for year in tick_text]
            range_x = [
                min_datetime - relativedelta(months=6),
                max_datetime + relativedelta(months=6),
            ]

        elif self.xaxis_tick_text_format == XAxisFormat.MONTH_YEAR.value:
            df = self.filtered_df.with_columns(
                pl.col(self.x_axis_column)
                .str.strptime(pl.Date, "%Y-%m-%d")
                .alias(self.x_axis_column)
            ).sort(self.x_axis_column)

            start_datetime = datetime(2024, 7, 1).date()
            latest_datetime = df[self.x_axis_column].max()
            tick_text = []
            current = start_datetime
            while current <= latest_datetime:
                tick_text.append(current.strftime("%b %Y"))
                current += relativedelta(months=1)

            tick_text_length = len(tick_text)
            total_tick_points = int((tick_text_length / 5) * 7)
            additional_tick_points = total_tick_points - tick_text_length
            last_current_tick_text = datetime.strptime(tick_text[-1], "%b %Y")

            for x in range(additional_tick_points):
                tick_text = tick_text + [
                    (last_current_tick_text + relativedelta(months=x + 1)).strftime(
                        "%b %Y"
                    )
                ]

            tick_values = [
                datetime.strptime(month_year, "%b %Y").replace(day=1)
                for month_year in tick_text
            ]

            range_x = [
                tick_values[0],
                tick_values[-1] + relativedelta(months=1),
            ]
            tick_text = replace_jun_jul_month_abbreviations(tick_text)

        elif self.xaxis_tick_text_format == XAxisFormat.MONTH_YEAR_MONTHLY_DATA.value:
            df = self.filtered_df.with_columns(
                pl.col(self.x_axis_column)
                .str.strptime(pl.Date, "%Y-%m-%d")
                .alias(self.x_axis_column)
            ).sort(self.x_axis_column)

            start_datetime = datetime(2024, 7, 1).date()
            latest_datetime = df[self.x_axis_column].max()
            extra_datetime = latest_datetime + relativedelta(months=1)

            tick_text = []
            current = start_datetime
            while current <= extra_datetime:
                tick_text.append(current.strftime("%b %Y"))
                current += relativedelta(months=1)

            tick_values = [
                datetime.strptime(month_year, "%b %Y").replace(day=1)
                for month_year in tick_text
            ]

            range_x = [
                tick_values[0],
                tick_values[-1],
            ]
            tick_text = replace_jun_jul_month_abbreviations(tick_text)
        elif self.xaxis_tick_text_format == XAxisFormat.FINANCIAL_QUARTER.value:
            tick_values = [1, 2, 3, 4]
            tick_text = [
                convert_financial_quarter_to_financial_quarter_text(quarter)
                for quarter in tick_values
            ]

            range_x = [0.5, 4.5]
        else:
            raise ValueError(
                f"Invalid xaxis_tick_text_format: {self.xaxis_tick_text_format}"
            )
        return tick_text, tick_values, range_x

    def _get_y_axis_range_max(self):
        """Get the y axis range maximum value to ensure there is an axis label greater than the
        maximum y value."""
        largest_number_of_weeks = self.filtered_df[self.y_axis_column].max()

        y_axis_max = largest_number_of_weeks + (0.3 * largest_number_of_weeks)
        return y_axis_max

    def _get_df_list_for_time_series(self) -> list[pl.DataFrame]:
        if self.trace_name_column is not None:
            df_list = [
                self.filtered_df.filter(pl.col(self.trace_name_column) == trace_name)
                for trace_name in self.trace_name_list
            ]
        else:
            df_list = [self.filtered_df]
        return df_list

    def _get_colour_list(self):
        """Returns a list of colours."""
        number_of_traces = len(self.trace_name_list)
        if number_of_traces == 2 and self.filled_traces_dict is None:
            colour_list = [
                AFAccessibleColours.DARK_BLUE.value,
                AFAccessibleColours.ORANGE.value,
            ]  # if 2 lines should use dark blue & orange as have highest contrast ratio
        else:
            colour_list = AFAccessibleColours.CATEGORICAL.value.copy()
        colour_shift_dict = (
            {"default": 0}
            if self.number_of_traces_colour_shift_dict is None
            else self.number_of_traces_colour_shift_dict
        )

        colour_shift_value = colour_shift_dict.get(
            number_of_traces, colour_shift_dict["default"]
        )
        if isinstance(colour_shift_value, list):
            return colour_shift_value  # list of colours
        while colour_shift_value > 0:
            colour_list.append(colour_list.pop(0))
            colour_shift_value -= 1
        return colour_list



"""
Function to add a prefix and suffix to a number and also show the number of decimal places
specified.
"""
import math
import decimal


def round_and_add_prefix_and_suffix(
    value, decimal_places=None, prefix="", suffix="", separator="-"
):
    """
    Rounds a number to the specified number of decimal places, for example rounds 12 to 2dp as
    12.00. Returns the formatted number as a string with and optional prefix and suffix.

    Args:
        value(float or int): number to format
        prefix(str,optional): prefix to append to the start (e.g. £). Defaults to no prefix.
        suffix(str,optional): suffix to append to the end (e.g. %). Defaults to no suffix.
        decimal_places(int, optional): If set, rounds formatted number to that many d.p.
            Defaults to None, which does not apply rounding.
        separator(str,optional): If set, will be returned for a NaN or None passed as
            value_to_format

    Returns:
        str: formatted number
    """
    if value is None or math.isnan(value):
        return separator

    if decimal_places is not None:
        decimal_value = decimal.Decimal(str(value))

        if decimal_places >= 0:
            value = decimal_value.quantize(
                decimal.Decimal(f"0.{'0' * decimal_places}"),
                rounding=decimal.ROUND_HALF_UP,
            )
        else:
            value = round(value, decimal_places)

    return f"{prefix}{value}{suffix}"



"""Choropleth map class"""

import polars as pl
import plotly.graph_objects as go
from dash import dcc

from gov_uk_dashboards.constants import DEFAULT_COLOURSCALE

from gov_uk_dashboards import colours
from gov_uk_dashboards.components.helpers.display_chart_or_table_with_header import (
    display_chart_or_table_with_header,
)


class ChoroplethMap:
    """Class for  generating choropleth map charts.
    Note: dataframe_function must contain columns: 'Region', 'Area_Code',
    discrete_category_column, column_to_plot, hover_data_list
    If discrete_category_column & discrete_category_order are None,
    the choropleth map will be a continuos one, otherwise discrete"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    # pylint: disable=too-few-public-methods
    def __init__(
        self,
        map_name: str,  # used for the id in the choropleth map and the data download button
        get_dataframe: callable,
        get_geos: callable,
        region: str,
        area_focus_level: str,
        column_to_plot: str,
        hover_header_list: list[str],
        hover_data_list: list[str],
        discrete_category_column: str = None,
        discrete_category_order: list[str] = None,
        legend_title_text: str = None,
        **choropleth_properties,
    ):
        self.map_name = map_name
        self.dataframe = get_dataframe()
        self.geographic_boundaries = get_geos()
        self.region = region
        self.area_focus_level = area_focus_level
        self.column_to_plot = column_to_plot
        self.hover_header_list = hover_header_list
        self.hover_data_list = hover_data_list

        self.discrete_category_column = discrete_category_column
        self.legend_title_text = legend_title_text
        self.discrete_category_order = discrete_category_order
        self.choropleth_properties = choropleth_properties

        self.fig = go.Figure()
        self.discrete_map = (
            self.discrete_category_column is not None
            and self.discrete_category_order is not None
        )
        if self.discrete_map:
            self.df_dict = self._get_dataframe_dict_by_category()
            self.colours_list = self._get_colour_list()

    def get_choropleth_map(self):
        """Creates and returns choropleth map chart for display on application.

        Returns:
            html.Div: Styled div containing title, subtile and chart.
        """
        self._update_fig()
        choropleth_map = dcc.Graph(
            id=f"{self.map_name}-choropleth",
            responsive=True,
            config={"topojsonURL": "/assets/topojson/", "displayModeBar": False},
            figure=self.fig,
            style={
                "height": "750px",
                "width": "100%",
            },  # height hard-coded so that map always displays within tab
        )
        return display_chart_or_table_with_header(
            choropleth_map, download_data_button_id=self.map_name
        )

    def _update_fig(self):
        self._add_traces()
        self._handle_missing_data()
        self._crop_to_region()
        self._remove_background_map()

    def _create_choropleth_trace(
        self,
        dataframe,
        colourscale,
        column_to_plot=None,
        is_missing_data=False,
        marker=None,
    ):
        # pylint: disable=too-many-positional-arguments
        """Return a chlorepleth trace set up for UK LAs, with optional configuration for markers"""
        if not column_to_plot:
            column_to_plot = self.column_to_plot
        if not is_missing_data:
            hovertemplate = (
                "<b>%{hovertext}</b><br>"
                + "<br>".join(
                    f"{header}: " + "%{customdata[" + str(i) + "]}"
                    for i, header in enumerate(self.hover_header_list)
                )
                + "<br><br><extra></extra>"
            )
        else:
            hovertemplate = (
                "<b>%{customdata[0]}</b><br>" + "%{hovertext}<extra></extra>"
            )

        return go.Choropleth(
            geojson=self.geographic_boundaries,
            featureidkey="properties.geo_id",
            locations=dataframe["Area_Code"],
            locationmode="geojson-id",
            z=dataframe[column_to_plot],
            hovertext=(
                dataframe[self.area_focus_level]
                if not is_missing_data
                else ["No data available"] * len(dataframe[self.area_focus_level])
            ),
            customdata=(
                dataframe[self.hover_data_list]
                if not is_missing_data
                else dataframe[[self.area_focus_level]]
            ),
            colorbar=self._get_color_bar(),
            hovertemplate=hovertemplate,
            marker=marker,
            marker_line_color=colours.GovUKColours.DARK_GREY.value,
            showscale=self._get_scale(),
            showlegend=self._get_legend(),
            colorscale=colourscale,
            name=self._get_trace_name(dataframe, is_missing_data),
            **self.choropleth_properties,
        )

    def _add_traces(self):
        if not self.discrete_map:
            self.fig.add_trace(
                self._create_choropleth_trace(self.dataframe, DEFAULT_COLOURSCALE)
            )
        else:
            for count, category in enumerate(self.discrete_category_order):
                if category not in self.df_dict:
                    df = self._create_df_for_empty_trace(category)
                else:
                    df = self.df_dict[category]
                colour = [
                    [0, self.colours_list[count % len(self.colours_list)]],
                    [1, self.colours_list[count % len(self.colours_list)]],
                ]
                self.fig.add_trace(
                    self._create_choropleth_trace(
                        df,
                        colour,
                    )
                )

    def _handle_missing_data(self):
        missing_data = self.dataframe.filter(pl.col(self.column_to_plot).is_null())
        missing_data = missing_data.with_columns(pl.lit(0).alias("data"))

        colour = [
            [0, colours.GovUKColours.MID_GREY.value],
            [1, colours.GovUKColours.MID_GREY.value],
        ]
        if not missing_data.is_empty():
            self.fig.add_trace(
                self._create_choropleth_trace(
                    missing_data,
                    colour,
                    column_to_plot="data",
                    is_missing_data=True,
                )
            )

    def _get_dataframe_dict_by_category(self):
        if self.region is not None and self.region != "England":
            self.dataframe = self.dataframe.filter(pl.col("Region") == self.region)
        else:
            self.dataframe = self.dataframe
        grouped_dfs_dict_keys_as_tuples = self.dataframe.partition_by(
            self.discrete_category_column, as_dict=True
        )
        grouped_dfs_dict = {
            key[0]: value for key, value in grouped_dfs_dict_keys_as_tuples.items()
        }
        return grouped_dfs_dict

    def _crop_to_region(self):
        self.fig.update_layout(
            {
                "margin": {"r": 0, "t": 0, "l": 0, "b": 0},
                "geo": {
                    "bgcolor": "rgba(0, 0, 0, 0)",
                    "fitbounds": "locations",
                },
                "coloraxis_colorbar_title_text": "",
                "dragmode": False,  # fixes and not zoomable
                "legend_x": 0.9,
                "legend_y": 0.95,
                "legend_title_text": self.legend_title_text,
            },
        )

    def _wrap_text(self, text, max_chars_per_line):
        words = text.split()
        lines = []
        current_line = ""

        for word in words:
            if len(current_line) + len(word) + 1 > max_chars_per_line:
                lines.append(current_line)
                current_line = word
            else:
                if current_line:
                    current_line += " "
                current_line += word

        if current_line:
            lines.append(current_line)

        return "<br>".join(lines)

    def _get_color_bar(self):
        if self.discrete_map:
            return None

        return {
            "title": self._wrap_text(self.legend_title_text, 10),
            "thickness": 20,
            "len": 0.8,
            "x": 0.9,
            "y": 0.5,
        }

    def _get_scale(self):
        return not self.discrete_map

    def _get_legend(self):
        return self.discrete_map

    def _get_trace_name(self, dataframe, missing_data=False):
        if self.discrete_category_column is None:
            return None

        if not missing_data:
            return dataframe[self.discrete_category_column][0]

        return "No data available"

    def _remove_background_map(self):
        self.fig.update_geos(
            center={"lat": 53, "lon": -2},
            visible=False,
            projection_type="mercator",
        )

    def _get_colour_list(self):
        """Amends colour list based on the number of categories"""
        colour_list = ["#217847", "#23BBBE", "#8CCE69", "#FFEA80"]

        if self.discrete_map and len(self.discrete_category_order) == 3:
            colour_list.pop(1)
        return colour_list

    def _create_df_for_empty_trace(self, category):
        """Method to create df where all columns are empty except for the category column to force
        all legend items to always appear"""
        columns = next(iter(self.df_dict.values())).columns
        data = {col: [None] for col in columns}
        data[self.discrete_category_column] = [category]
        return pl.DataFrame(data)



"""stacked_barchart function"""

import json
import math
from typing import Optional
from dash import html
import polars as pl

import plotly.graph_objects as go

from gov_uk_dashboards.components.helpers.get_chart_for_download import (
    get_chart_for_download,
)
from gov_uk_dashboards.constants import (
    CHART_LABEL_FONT_SIZE,
    CUSTOM_DATA,
    DATE_VALID,
    FINANCIAL_YEAR_ENDING,
    HOVER_TEXT_HEADERS,
    LEGEND_SPACING,
    MAIN_TITLE,
    MEASURE,
    SUBTITLE,
    VALUE,
)
from gov_uk_dashboards.colours import AFAccessibleColours
from gov_uk_dashboards.components.helpers.display_chart_or_table_with_header import (
    display_chart_or_table_with_header,
)
from gov_uk_dashboards.components.helpers.generate_dash_graph_from_figure import (
    generate_dash_graph_from_figure,
)
from gov_uk_dashboards.components.helpers.plotting_helper_functions import (
    get_legend_configuration,
)
from gov_uk_dashboards.components.plotly.enums import (
    HoverDataByTrace,
    TitleDataStructure,
    XAxisFormat,
)
from gov_uk_dashboards.formatting.human_readable import format_as_human_readable

from gov_uk_dashboards.components.helpers.update_layout_bgcolor_margin import (
    update_layout_bgcolor_margin,
)


class StackedBarChart:
    """Class for use in generating stacked bar charts."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable = too-many-positional-arguments
    def __init__(
        self,
        title_data: TitleDataStructure,
        y_axis_column: str,
        hover_data: HoverDataByTrace,
        df: pl.DataFrame,
        trace_name_list: list[str],
        trace_name_column: Optional[str] = None,
        xaxis_tick_text_format: XAxisFormat = XAxisFormat.YEAR.value,
        line_trace_name: Optional[str] = None,
        x_axis_column=DATE_VALID,
        x_unified_hovermode: Optional[bool] = False,
        hover_distance: Optional[int] = 1,
        download_chart_button_id: Optional[str] = None,
        download_data_button_id: Optional[str] = None,
        total_trace_name: Optional[str] = None,
    ):
        """Initializes the StackedBarChart instance.
        To display the chart, call the `get_stacked_bar_chart()` method.

        Args:
            title_data (TitleDataStructure): Data structure containing the chart title information.
            y_column (str): The column name representing the Y-axis data.
            hover_data (HoverDataByTrace): Data structure for hover information.
            df (pl.DataFrame): The dataset for the chart.
            trace_name_list (list[str]): List of trace names for the stacked bars.
            trace_name_column (Optional[str], optional): Column name representing trace categories,
                if applicable. Defaults to None.
            xaxis_tick_text_format (XAxisFormat, optional): Format for X-axis tick labels.
                Defaults to XAxisFormat.YEAR.value.
            line_trace_name (Optional[str], optional): Name for an optional line trace overlay,
                must be in MEASURE column of df, line_trace_name will display in legend.
                Defaults to None.
            x_axis_column (_type_, optional): The column used for the X-axis values.
                Defaults to DATE_VALID.
            download_chart_button_id (Optional[str], optional): ID for the chart download button,
                if applicable. Defaults to None.
            download_data_button_id (Optional[str], optional): ID for the data download button, if
                applicable. Defaults to None.
            total_trace_name (Optional[str], optional): Name for an optional total to be added to
                bottom of hover text, must be in MEASURE column of df, line_trace_name will display
                in legend. Defaults to None.
        """
        self.title_data = title_data
        self.y_axis_column = y_axis_column
        self.hover_data = hover_data
        self.df = df
        self.trace_name_list = trace_name_list
        self.trace_name_column = trace_name_column
        self.xaxis_tick_text_format = xaxis_tick_text_format
        self.line_trace_name = line_trace_name
        self.x_axis_column = x_axis_column
        self.x_unified_hovermode = x_unified_hovermode
        self.hover_distance = hover_distance
        self.download_chart_button_id = download_chart_button_id
        self.download_data_button_id = download_data_button_id
        self.total_trace_name = total_trace_name
        self.fig = self.create_stacked_bar_chart()

    def get_stacked_bar_chart(self) -> html.Div:
        """Creates and returns stacked bar chart for display on application.

        Returns:
            html.Div: Styled div containing title, subtile and chart.
        """
        graph_name = self.title_data[MAIN_TITLE].replace(" ", "-")
        return display_chart_or_table_with_header(
            generate_dash_graph_from_figure(
                self.fig,
                graph_name,
                class_name="default-cursor-graph non-interactive-legend-cursor",
            ),
            self.title_data[MAIN_TITLE],
            self.title_data[SUBTITLE],
            self.download_chart_button_id,
            self.download_data_button_id,
        )

    def is_json_serializable(self, value):
        "Determines whether value can be converted to json format."
        try:
            json.dumps(value)
            return True
        except (TypeError, OverflowError):
            return False

    def to_dict(self):
        "Converts class attributes to json format."
        result = {}
        for k, v in self.__dict__.items():
            if self.is_json_serializable(v):
                result[k] = v
            elif isinstance(v, pl.DataFrame):
                result[k] = {"_type": "polars_df", "data": v.to_dicts()}
            elif hasattr(v, "to_dict"):
                result[k] = {"_type": "custom", "data": v.to_dict()}
            else:
                result[k] = f"<<non-serializable: {type(v).__name__}>>"
        return result

    @classmethod
    def from_dict(cls, data):
        "Creates a class instance from dict of attributes."
        restored = {}
        for k, v in data.items():
            if isinstance(v, dict) and "_type" in v:
                if v["_type"] == "polars_df":
                    restored[k] = pl.DataFrame(v["data"])
                elif v["_type"] == "custom":
                    # optionally restore known nested types here
                    pass
            else:
                restored[k] = v
        return cls(**restored)

    def get_stacked_bar_chart_for_download(self):
        """Return fig with title and subtitle for download as png"""
        return get_chart_for_download(self, self.create_stacked_bar_chart())

    def create_stacked_bar_chart(
        self,
    ):
        """generates a stacked bar chart"""
        # pylint: disable=too-many-locals

        fig = go.Figure()
        if self.total_trace_name is not None:
            df = self.df.filter(pl.col(MEASURE) == self.total_trace_name)

            fig.add_trace(
                go.Scatter(
                    x=df[self.x_axis_column],
                    y=df[self.y_axis_column],
                    customdata=self._get_custom_data(self.total_trace_name, df),
                    mode="markers",
                    marker={"color": "white", "opacity": 0},
                    name=self.total_trace_name + LEGEND_SPACING,
                    hovertemplate="Total income: %{customdata[0]}<extra></extra>",
                    showlegend=False,
                    hoverinfo="skip",
                    legendrank=1,
                )
            )
        colour_list = (
            AFAccessibleColours.CATEGORICAL.value
            if len(self.trace_name_list) != 2
            else [
                AFAccessibleColours.DARK_BLUE.value,
                AFAccessibleColours.ORANGE.value,
            ]  # if 2 lines should use dark blue & orange as have highest contrast ratio
        )
        for _, (df, trace_name, colour) in enumerate(
            zip(
                self._get_df_list_for_bar_chart(),
                self.trace_name_list,
                colour_list,
            )
        ):
            fig.add_trace(
                self.create_bar_chart_trace(
                    df.sort(self.x_axis_column),
                    trace_name,
                    hover_label=None,
                    colour=colour,
                )
            )

        if self.line_trace_name is not None:
            colour = AFAccessibleColours.CATEGORICAL.value[len(self.trace_name_list)]
            df = self.df.filter(pl.col(MEASURE) == self.line_trace_name)

            fig.add_trace(
                go.Scatter(
                    x=df[FINANCIAL_YEAR_ENDING],
                    y=df[VALUE],
                    customdata=self._get_custom_data(self.line_trace_name, df),
                    mode="lines",
                    line={"color": colour, "width": 3},
                    name=self.line_trace_name + LEGEND_SPACING,
                    hovertemplate=self._get_hover_template(self.line_trace_name),
                    legendrank=99999,  # a high number to ensure it is bottom of the legend
                )
            )

        max_y, min_y, tickvals, ticktext = _get_y_range_tickvals_and_ticktext(
            self.df, "£", self.trace_name_list
        )
        update_layout_bgcolor_margin(fig, "#FFFFFF")

        fig.update_layout(
            legend=get_legend_configuration(reverse_order=True),
            font={"size": CHART_LABEL_FONT_SIZE},
            yaxis={
                "range": [min_y * 1.1, max_y * 1.1],
                "tickmode": "array",
                "tickvals": tickvals,
                "ticktext": ticktext,
            },
            showlegend=True,
            barmode="relative",
            xaxis={"categoryorder": "category ascending"},
            ## copied from timeseries
            hovermode="x unified" if self.x_unified_hovermode is True else "closest",
            hoverdistance=self.hover_distance,  # Increase distance to simulate hover 'always on'
        )
        return fig

    def create_bar_chart_trace(
        self,
        df: pl.DataFrame,
        trace_name: str,
        hover_label: dict[str, str],
        colour: str,
    ):
        """Creates a trace for the plot.

        Args:
            df (pl.DataFrame): Dataframe to use to create trace. Must contain x and y columns,
            and columns defined in self.hover_data[CUSTOM_DATA].
            trace_name (str): Name of trace.
            hover_label (dict[str,str]): Properties for hoverlabel parameter.
            colour (str): Colour for bar.
        """

        return go.Bar(
            x=df[self.x_axis_column],
            y=df[self.y_axis_column],
            name=trace_name + LEGEND_SPACING,
            hovertemplate=[
                self._get_hover_template(trace_name) for i in range(df.shape[0])
            ],
            customdata=self._get_custom_data(trace_name, df),
            hoverlabel=hover_label,
            marker={"color": colour},
        )

    def _get_hover_template(self, trace_name):
        if self.x_unified_hovermode is True:
            return f"{trace_name}: " + "%{customdata[0]}<extra></extra>"
        hover_text_headers = self.hover_data[trace_name][HOVER_TEXT_HEADERS]
        hover_template = (
            f"{trace_name}<br>"
            f"{hover_text_headers[0]}"
            ": %{customdata[0]}<br>"
            f"{hover_text_headers[1]}"
            ": %{customdata[1]}<extra></extra>"
        )
        return hover_template

    def _get_custom_data(self, trace_name, df):
        customdata = df[self.hover_data[trace_name][CUSTOM_DATA]]
        return customdata

    def _get_df_list_for_bar_chart(self) -> list[pl.DataFrame]:
        if self.trace_name_column is not None:
            df_list = [
                self.df.filter(pl.col(self.trace_name_column) == trace_name)
                for trace_name in self.trace_name_list
            ]
        else:
            df_list = [self.df]
        return df_list


def _get_y_range_tickvals_and_ticktext(
    dataframe: pl.DataFrame, tick_prefix: str, yaxis_with_values: list[str]
):
    barchart_df = dataframe.pivot(
        index=FINANCIAL_YEAR_ENDING, columns=MEASURE, values=VALUE
    )
    positive_sum = sum(
        pl.when(pl.col(col) > 0).then(pl.col(col)).otherwise(0)
        for col in yaxis_with_values
    )
    negative_sum = sum(
        pl.when(pl.col(col) < 0).then(pl.col(col)).otherwise(0)
        for col in yaxis_with_values
    )
    barchart_df = barchart_df.with_columns(positive_sum.alias("Positive sum"))
    barchart_df = barchart_df.with_columns(negative_sum.alias("Negative sum"))
    maxy = barchart_df.select([pl.col("Positive sum").max()]).item()
    miny = barchart_df.select([pl.col("Negative sum").min()]).item()
    tickvals = _generate_tickvals(maxy, miny)
    ticktext = [format_as_human_readable(val, prefix=tick_prefix) for val in tickvals]
    return tickvals[-1], tickvals[0], tickvals, ticktext


def _generate_tickvals(maxy, miny):
    range_size = maxy - miny

    # Determine the order of magnitude of the range
    order = int(math.log10(range_size))

    # Start with an initial step size
    step_size = 10**order

    # Calculate the number of ticks based on the step size
    num_ticks = math.ceil(range_size / step_size)

    # Adjust step size to ensure the number of ticks is between 6 and 10
    while num_ticks < 6 or num_ticks > 10:
        if num_ticks < 6:  # Too few ticks -> decrease step size
            step_size /= 2
        elif num_ticks > 10:  # Too many ticks -> increase step size
            step_size *= 2
        num_ticks = math.ceil(range_size / step_size)

    # Adjust the start and end of the range to align with the step size
    start = math.floor(miny / step_size) * step_size
    end = math.ceil(maxy / step_size) * step_size

    # Generate tick values
    tickvals = list(range(int(start), int(end) + 1, int(step_size)))
    return tickvals
